package test;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class SaxParser extends DefaultHandler {

  private static final int DEFAULT_LENGTH = 64;

  private final Map<String, List<QuestData>> quests;   // zone -> list of quests
  private final Set<String> ignoredScopes;
  private final Set<String> ignoredQuests;

  private final SaxPath path;
  private final StringBuilder buf;
  private QuestData data;

  private boolean inQuests;
  private boolean inQuest;
  private boolean inQuestId;
  private boolean inRepeatable;
  private boolean inName;
  private boolean inZone;
  private boolean inFaction;
  private boolean inScope;
  private boolean inScene;

  // Constructors

  public SaxParser(Map<String, List<QuestData>> pQuests) {
    quests = pQuests;
    path = new SaxPath();
    buf = new StringBuilder(DEFAULT_LENGTH);

    ignoredScopes = new HashSet<>();
    ignoredScopes.add("Instant Adventure");
    ignoredScopes.add("Guild");
    ignoredScopes.add("Raid");
    ignoredScopes.add("Weekly");
    ignoredScopes.add("PvP");
    ignoredScopes.add("Ancient Wardstones");
    ignoredScopes.add("Soul");

    ignoredQuests = new HashSet<>();
    /*
    ignoredQuests.add("Golden Embers");   // These apparently are all repeatable!
    ignoredQuests.add("Hammer and Flame");
    ignoredQuests.add("The Blight");
    ignoredQuests.add("Waves of Madness");
    ignoredQuests.add("Infernal Awakening");
    ignoredQuests.add("Into the Depths");
    ignoredQuests.add("A Clash of Power");
    ignoredQuests.add("Summer Swarms");   // ???
    ignoredQuests.add("The Grim Harvest");   // ???
    ignoredQuests.add("Parade Day!");   // ???
    ignoredQuests.add("Visions of Future Past");   // ???

    ignoredQuests.add("The Embers of Silverwood");   // Silverwood
    ignoredQuests.add("Arboreal Assault");
    ignoredQuests.add("The Forest Fire");
    ignoredQuests.add("Crusade Against House Aelfwar");

    ignoredQuests.add("Battle for Freemarch");   // Freemarch
    ignoredQuests.add("The Crushing Deep");
    ignoredQuests.add("Fall Of The Hammer");
    ignoredQuests.add("The Hand of Death");
    ignoredQuests.add("The Ebb of the Tidelord");

    ignoredQuests.add("The Purging of Gloamwood Pines");   // Gloamwood
    ignoredQuests.add("The Wild Plague");
    ignoredQuests.add("Forged in Rage");
    ignoredQuests.add("Descending Night");
    ignoredQuests.add("From the Deeps");

    ignoredQuests.add("Field Of Screams");   // Stonefield
    ignoredQuests.add("The Stone Titan");
    ignoredQuests.add("Undead Burden");
    
    ignoredQuests.add("The Bone Dance");   // Scarlet Gorge
    ignoredQuests.add("The Crucible");
    ignoredQuests.add("Shambling Hordes");
    ignoredQuests.add("Fires of Suffering");
    ignoredQuests.add("Systematic Deconstruction");
    ignoredQuests.add("Greed in the Gorge");
    ignoredQuests.add("A Royal Raid");

    ignoredQuests.add("Winged Horror");   // Scarwood Reach
    ignoredQuests.add("Explosive Life");
    ignoredQuests.add("Hard Knocks");
    ignoredQuests.add("Double Danger");
    ignoredQuests.add("The Witching Hour");
    // TODO: events from other zones
    */
  }

  // Methods

  private boolean isIncluded(QuestData quest) {
    if (quest.hasProperty("Scope")) {
      String scope = quest.getProperty("Scope");
      for (String str : ignoredScopes) {
        if (scope.contains(str)) {
          return false;
        }
      }
    }
    String name = quest.getProperty("Name");
    return !name.startsWith("Workorder:") && !ignoredQuests.contains(name) && quest.hasProperty("Zone")
        && (!quest.hasProperty("Repeatable") || quest.getProperty("Repeatable").equals("Never"));
  }

  @Override
  public void characters(char[] ch, int start, int length) throws SAXException {
    if (inQuests && inQuest) {
      if (inQuestId || inZone || inFaction || inName || inScope || inRepeatable || inScene) {
        buf.append(ch, start, length);
      }
    }
  }

  @Override
  public void endElement(String uri, String localName, String qName) throws SAXException {
    path.pop();

    if (qName.equals("English") && inName) {
      inName = false;
      data.addProperty("Name", buf.toString());
    }
    else if (qName.equals("Scope") && inScope) {
      inScope = false;
      data.appendProperty("Scope", buf.toString());
    }
    else if (qName.equals("QuestId") && inQuestId) {
      inQuestId = false;
      data.addProperty("QuestId", buf.toString());
    }
    else if (qName.equals("Zone") && inZone) {
      inZone = false;
      data.appendProperty("Zone", buf.toString());
    }
    else if (qName.equals("Faction") && inFaction) {
      inFaction = false;
      data.appendProperty("Faction", buf.toString());
    }
    else if (qName.equals("Scene")) {
      inScene = false;
      data.appendProperty("Scene", buf.toString());
    }
    else if (qName.equals("Repeatable")) {
      inRepeatable = false;
      data.appendProperty("Repeatable", buf.toString());
    }
    else if (qName.equals("Quest")) {
      inQuest = false;
      if (isIncluded(data)) {
        String[] zones = data.getProperty("Zone").split(",");
        for (String zone : zones) {
          List<QuestData> zoneQuests = quests.get(zone);
          if (zoneQuests == null) {
            zoneQuests = new ArrayList<>();
            quests.put(zone, zoneQuests);
          }
          zoneQuests.add(data);
        }
      }
    }
    else if (qName.equals("Quests")) {
      inQuests = false;
    }
  }

  @Override
  public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
    path.push(qName);

    if (qName.equals("English") && path.equals("/Quests/Quest/Name/English")) {
      inName = true;
      buf.setLength(0);
    }
    else if (qName.equals("Scope") && path.equals("/Quests/Quest/Scope")) {
      inScope = true;
      buf.setLength(0);
    }
    else if (qName.equals("QuestId") && path.equals("/Quests/Quest/QuestId")) {
      inQuestId = true;
      buf.setLength(0);
    }
    else if (qName.equals("Zone") && path.equals("/Quests/Quest/Zone")) {
      inZone = true;
      buf.setLength(0);
    }
    else if (qName.equals("Faction") && path.equals("/Quests/Quest/Faction")) {
      inFaction = true;
      buf.setLength(0);
    }
    else if (qName.equals("Scene")) {
      inScene = true;
      buf.setLength(0);
    }
    else if (qName.equals("Repeatable")) {
      inRepeatable = true;
      buf.setLength(0);
    }
    else if (qName.equals("Quest")) {
      inQuest = true;
      data = new QuestData();
    }
    else if (qName.equals("Quests")) {
      inQuests = true;
    }
  }

}
